from .base_component import *


class JAAQLComponent(BaseComponent):

    def test_jaaql(self):
        pass

    # test submit, test submitting incorrect arguments incorrect types, incorrect url, rate limiting